/**
 Simulation of a medeival cannon. It takes into account gravity
 and air resistance.
 
 This simulation fires a cannon balls every five dgrees between 10 and 
 50 degrees. Once the cannon has fired it will not be ready to fire again
 until the cannon ball has hit the ground.
 
 Vectors are used to hold all positions, velocities and accelerations
 (gravity and air resistance).
 
 
 Used in the blog "Cannon Fire" at http://lagers.org.uk/blog/
 
 created 2016 Peter Lager
 */
PImage bg;

ArrayList<PVector> track = new ArrayList<PVector>();

int currTime, prevTime;    // milliseconds
float elapsedTime;         // seconds
float speed = 300;         // cannon-shot speed

int time_to_fire;          // The next time to fire the cannon 

// Cannon data
PShape cannon;              // graphic
boolean cannonReady = true; // ready to fire
int cannon_angle;           // degrees
PVector cannon_pos;

// SHot data
PVector ball_pos, ball_vel;
// Forces (gravity and air resistance)
PVector gravity, air;

void setup() {
  size(800, 200, P2D);
  bg = loadImage("bg2.jpg");
  cursor(CROSS);
  cannon_pos = new PVector(80, height - 50);
  ball_pos = new PVector();
  ball_vel = new PVector();
  gravity = new PVector(0, 100);
  air = new PVector(-40, 0);
  cannon = createCannon();
  noStroke();
  // Initialise the cannon and shot
  cannon_angle = initCannon(0, 10);
  // Set the timer variables (should be last thing in setup)
  currTime = prevTime = millis();
  // The first shot will be in 500ms
  time_to_fire = millis() + 500;
}

// Initialise the cannon and shot. Notice that we use negative angles 
// to rotate the barrel counter-clockwise.
int initCannon(int lastAngle, int newAngle) {
  // Remove existing angle from cannon
  cannon.getChild(0).rotate(-radians(-lastAngle));
  // Apply new angle to cannon
  cannon.getChild(0).rotate(radians(-newAngle));
  // Initialise ball position
  ball_pos.set(cannon_pos);
  // Calculate the ball's velocity vector
  float angle = radians(newAngle);
  ball_vel.set(speed * cos(-angle), speed * sin(-angle));
  return newAngle;
}

void draw() {
  // Calculate the elapsed time since the last frame
  currTime = millis();
  elapsedTime = (currTime - prevTime) / 1000.0f;
  prevTime = currTime;  // set it up for the next frame

  background(bg);
  if (!cannonReady) {
    // Update velocity due to affect of gravity and air resistance
    ball_vel.add(PVector.mult(gravity, elapsedTime));
    ball_vel.add(PVector.mult(air, elapsedTime));
    // Fix to prevent cannon ball moving towards the cannon under the
    // influence of air resistance
    ball_vel.x = max(ball_vel.x, 0);
    // Update the cannon ball's position
    ball_pos.add(PVector.mult(ball_vel, elapsedTime));
    // Record the cannon ball position so we can display its trajectory
    track.add(ball_pos.copy());
    // display cannon ball
    fill(64);
    ellipse(ball_pos.x, ball_pos.y, 6, 6);
    // See if the shot has hit the ground
    if (ball_pos.y > height - 32) {
      cannonReady = true;
      // Prepare to fire next shot in 1.2 seconds
      time_to_fire = millis() + 1200;
    }
  } else {
    // Cannon is ready to fire so see if it is time to do it.
    if (millis() > time_to_fire) {
      // Calculate the next angle to elevate the cannon
      int angle = cannon_angle + 5;
      // If we are more than 50 degrees start back at 
      // 10 degrees and clear the tracks
      if (angle > 50) {
        angle = 10;
        track.clear();
      }
      // Set the new cannon angle and reset the shoat
      cannon_angle = initCannon(cannon_angle, angle);
      cannonReady = false;
    }
  }
  // Diplsay the cannon ball trajectories as red dots
  fill(255, 0, 0);
  for (PVector tp : track)
    ellipse(tp.x, tp.y, 1.5, 1.5);
  // Display the cannon
  pushMatrix();
  translate(cannon_pos.x, cannon_pos.y);
  shape(cannon);
  popMatrix();
}